/**
 * Copyright (C) 2022 Igalia S.L. <info@igalia.com>
 *   Author: Fernando Jimenez Moreno <fjimenez@igalia.com>
 *   Author: Zan Dobersek <zdobersek@igalia.com>
 *   Author: Imanol Fernandez <ifernandez@igalia.com>
 *   Author: Loïc Le Page <llepage@igalia.com>
 *   Author: Jani Hautakangas <jani@igalia.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

package com.wpe.wpe;

import android.content.Context;
import android.content.Intent;
import android.os.Looper;
import android.os.ParcelFileDescriptor;
import android.util.Log;

import androidx.annotation.Keep;
import androidx.annotation.NonNull;
import androidx.annotation.Nullable;
import androidx.annotation.UiThread;
import androidx.annotation.WorkerThread;

import com.wpe.wpe.services.WPEServiceConnection;
import com.wpe.wpe.services.WPEServiceConnectionListener;

import java.io.File;

@UiThread
public final class Browser {
    private static final String LOGTAG = "WPEBrowser";

    static { System.loadLibrary("WPEAndroidBrowser"); }

    protected static native void startNativeLooper();
    private static native void setupNativeEnvironment(@NonNull String[] envStringsArray);
    private native void nativeInit(@NonNull String dataDir, @NonNull String cacheDir);
    private native void nativeShut();

    private static final Browser singleton = new Browser();

    public static @NonNull Browser getInstance() { return singleton; }

    private Browser() { Log.v(LOGTAG, "WPE Browser creation"); }

    private Context applicationContext = null;

    public @Nullable Context getApplicationContext() { return applicationContext; }

    private LooperHelperThread looperHelperThread = null;

    public void initialize(@NonNull Context context) {
        if (applicationContext == null) {
            applicationContext = context.getApplicationContext();
            setupNativeEnvironment(
                new String[] {"GIO_EXTRA_MODULES", new File(context.getFilesDir(), "gio").getAbsolutePath()});
            nativeInit(context.getDataDir().getAbsolutePath(), context.getCacheDir().getAbsolutePath());
            looperHelperThread = new LooperHelperThread();
        }
    }

    @Override
    protected void finalize() throws Throwable {
        super.finalize();
        nativeShut();
    }

    protected final AuxiliaryProcessesContainer auxiliaryProcesses = new AuxiliaryProcessesContainer();

    private final WPEServiceConnectionListener serviceConnectionDelegate = new WPEServiceConnectionListener() {
        public void onCleanExit(@NonNull WPEServiceConnection connection) {
            Log.d(LOGTAG, "onCleanExit for process: " + connection.getPid());
            auxiliaryProcesses.unregister(connection.getPid());
        }

        public void onServiceDisconnected(@NonNull WPEServiceConnection connection) {
            Log.d(LOGTAG, "onServiceDisconnected for process: " + connection.getPid());
            // Unbind the service to prevent restarting on crash,
            // wpewebkit will restart the auxiliary process if needed
            auxiliaryProcesses.unregister(connection.getPid());
        }
    };

    /**
     * This method is called directly from WebKit when a new auxiliary process needs to be created.
     * <p>In order to safeguard the rest of the system and allow the application to remain responsive even if the
     * user had loaded web page that infinite loops or otherwise hangs, the modern incarnation of WebKit uses
     * multi-process architecture. Web pages are loaded in its own WebContent process. Multiple WebContent processes
     * can share a browsing session, which lives in a shared network process. In addition to handling all network
     * accesses, this process is also responsible for managing the disk cache and Web APIs that allow websites to
     * store structured data such as Web Storage API and IndexedDB API.</p>
     * <p>Because a WebContent process can Just-in-Time compile arbitrary JavaScript code loaded from the internet,
     * meaning that it can write to memory that gets executed, this process is tightly sandboxed. It does not have
     * access to any file system unless the user grants an access, and it does not have direct access to the
     * underlying operating system’s clipboard, microphone, or video camera even though there are Web APIs that grant
     * access to those features. Instead, UI process brokers such requests.</p>
     * <p>Given that Android forbids the fork syscall on non-rooted devices, we cannot directly spawn child processes
     * . Instead we use Android Services to host the logic of WebKit's auxiliary processes.</p>
     * <p>The life cycle of all WebKit's auxiliary processes is managed by WebKit itself. We only proxy requests to
     * spawn and terminate these processes/services.</p>
     *
     * FIXME: except for the case where Android decides to kill a Service. In that case we need to notify WebKit (and
     * wait for WebKit to spawn the Service again).
     *
     * @param pid The process identifier. This value is generated by WebKit and does not correspond to the actual
     * system pid.
     * @param type The type of service to launch. It can be Web (0) or Network (1).
     * @param fd File descriptor used by WebKit for IPC.
     */
    @Keep
    @WorkerThread
    public void launchProcess(long pid, int type, int fd) {
        ProcessType processType = ProcessType.fromValue(type);
        Log.d(LOGTAG, "launchProcess " + processType.name() + " (pid: " + pid + ", fd: " + fd + ")");

        int processSlot = auxiliaryProcesses.getFirstAvailableSlot(processType);
        try {
            Class<?> serviceClass =
                Class.forName("com.wpe.wpe.services.WPEServices$" + processType.name() + "Service" + processSlot);
            ParcelFileDescriptor parcelFd = ParcelFileDescriptor.adoptFd(fd);

            Log.v(LOGTAG, "Launching service: " + processType.name());
            Intent intent = new Intent(applicationContext, serviceClass);

            WPEServiceConnection serviceConnection =
                new WPEServiceConnection(pid, processType, parcelFd, serviceConnectionDelegate);
            applicationContext.bindService(intent, serviceConnection,
                                           Context.BIND_AUTO_CREATE | Context.BIND_IMPORTANT);
            auxiliaryProcesses.register(pid, serviceConnection);
        } catch (Exception e) {
            Log.e(LOGTAG, "Cannot launch auxiliary process", e);
        }
    }

    /**
     * Terminate the Service hosting the logic for a WebKit auxiliary process that matches the given pid.
     * @param pid The process identifier. This value is generated by WebKit and does not correspond to the actual
     * system pid.
     */
    @Keep
    @WorkerThread
    public void terminateProcess(long pid) {
        auxiliaryProcesses.unregister(pid);
    }

    /**
     * A sideline thread that allows Java-based Looper execution from native code.
     */
    private static final class LooperHelperThread {
        private boolean isInitialized = false;

        private final Thread thread = new Thread(() -> {
            Looper.prepare();
            startNativeLooper();

            synchronized (LooperHelperThread.this) {
                isInitialized = true;
                LooperHelperThread.this.notifyAll();
            }

            Looper.loop();
        });

        LooperHelperThread() {
            thread.start();

            synchronized (this) {
                while (!isInitialized) {
                    try {
                        wait();
                    } catch (InterruptedException e) {
                        Log.d(LOGTAG, "LooperHelperThread has been interrupted");
                    }
                }
            }
        }
    }
}
